# Copyright Metaswitch Networks - Highly Confidential Material
"""This script is a POC for running a spirent test case or set of test cases automatically on one of our 1P ASE vm's."""

import json
import logging
import os

from landslide_test_runner.landslide_utils import LandslideTC
from landslide_test_runner.populate_credentials import (
    clear_credentials,
    populate_credentials,
)
from landslide_test_runner.update_nipap import deactivate_all_ips
from landslide_test_runner.yamlparse import (
    Sut,
    TestSession,
    get_config,
    get_session_id,
    parse_arguments,
)

logging.root.setLevel("INFO")
logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

# ---------------------
# Response Code
successfulRespCodeOK = 200
successfulRespCode = 201
result_dir = os.getcwd()
# Change user login to a generic ci value


def set_up_session(config, session):
    """Get all the session config set up before running."""
    populate_credentials()  # Get login details.
    sut_created = False
    user = config.identifier
    sut_created = config.check_test_config(user)
    login = config.spirent_configuration.get_login_details()
    landslide_session = create_landslide_session(session, login, config.sut)
    name = session.test_session_name
    return (landslide_session, name, sut_created)


def cleanup_session(config, sut_created):
    """Clean up resources from test session if the call to start the test fails."""
    logging.info("Cleaning up...")
    user = config.identifier
    deactivate_all_ips(user)
    if sut_created:
        config.sut.delete_sut()
    clear_credentials()


def execute_sessions(config_filename: str, action: str):
    """Get test sessions, check the input config is reasonable, and run them on Landslide."""
    config = get_config(config_filename)

    interface = "eth2"
    test_sessions = config.test_session_list
    for session in test_sessions:
        if action == "start":
            start_session(config, session)

        elif action == "monitor":
            monitor_session(config, session)

        elif action == "stop":
            stop_session(config, session, result_dir, interface)

        else:
            run_session_and_get_results(config, session, result_dir, interface)


def create_landslide_session(test_session: TestSession, user_details, sut: Sut):
    """Set up instance of landslide tc using test parameters."""
    test_session_name = test_session.test_session_name
    logging.info(
        "Running Test Session: %s with the following data:",
        test_session_name,
    )
    logging.info("YAML INPUT: %s", test_session)

    landslide_json = json.dumps(test_session.as_json(sut))
    logging.info(landslide_json)
    landslide_session = LandslideTC(landslide_json, test_session_name, user_details)
    return landslide_session


def start_session(config, session):
    """Starts test then returns."""
    # define the interface for the pcap to be downloaded
    (landslide_session, name, sut_created) = set_up_session(config, session)
    tas_ip = config.spirent_configuration.tas_ip
    landslide_session.id = get_session_id(name, tas_ip)

    ts = landslide_session

    response = ts.startTc(tas_ip)
    if response != successfulRespCode:
        logger.info("\n Failed to start test session - %s : %s", ts.nodeName, response)
        cleanup_session(config, sut_created)
    else:
        logger.info(
            "\n Successfully started test session - %s : %s", ts.nodeName, response
        )


def run_session_and_get_results(config, session, logdir, interface):
    """Run test session, wait until finished then return results."""
    (landslide_session, name, sut_created) = set_up_session(config, session)
    tas_ip = config.spirent_configuration.tas_ip
    landslide_session.id = get_session_id(name, tas_ip)
    try:
        ts = landslide_session
        response = ts.execute_testSession_and_store_results(tas_ip, logdir, interface)
        if response != successfulRespCode:
            logger.info(
                "\n Failed to start test session - %s : %s", ts.nodeName, response
            )
            return False
        else:
            logger.info(
                "\n Successfully started test session - %s : %s", ts.nodeName, response
            )
    finally:
        cleanup_session(config, sut_created)


def monitor_session(config, session):
    """Check the state of the test session."""
    get_session = set_up_session(config, session)
    ts = get_session[0]
    name = get_session[1]
    tas_ip = config.spirent_configuration.tas_ip
    logging.info("Retrieving test session id")
    ts.id = get_session_id(name, tas_ip)
    logging.info("Checking the status of test with id: %s", ts.id)
    tsStatus = ts.getTCStatus(tas_ip)
    logger.info("Test session status : %s ", (tsStatus))


def stop_session(config, session, logdir, interface):
    """Stop the test session and get our results."""
    (landslide_session, name, sut_created) = set_up_session(config, session)
    tas_ip = config.spirent_configuration.tas_ip
    landslide_session.id = get_session_id(name, tas_ip)
    ts = landslide_session
    try:
        logging.info("Stopping test session with id: %s", ts.id)
        response = ts.stopCmdSeqTc_and_getResults(tas_ip)
        logger.info("---------- %s : %s", ts.nodeName, response)
        if response != successfulRespCode:
            logger.info(
                "\n Failed to stop test session - %s : %s", ts.nodeName, response
            )
            return False
        else:
            ts.downloadPcap(tas_ip, interface, logdir)  # capture pcap
            ts.getTcExecSummaryStats(tas_ip, logdir)  # capture summary stats
    except KeyError:
        logger.info("Test could not be found or is no longer running.")
    finally:
        cleanup_session(config, sut_created)


def run():
    """Entry point for the test run calls."""
    logging.basicConfig(level=logging.DEBUG)
    arguments = parse_arguments()
    config_filename = arguments.test_config

    if arguments.start:
        execute_sessions(config_filename, "start")
    elif arguments.stop:
        execute_sessions(config_filename, "stop")
    elif arguments.monitor:
        execute_sessions(config_filename, "monitor")
    else:
        execute_sessions(config_filename, "run")


if __name__ == "__main__":
    run()
