#!/usr/bin/env python3
# Copyright Metaswitch Networks - Highly Confidential Material
"""Landslide Testcase execution utils."""
__copyright__ = "# Copyright (c) 2020 by Affirmed Networks, Inc. All rights reserved."
__author__ = """Kavita Kulkarni <kavita_kulkarni@affirmednetworks.com>
Umair Afzal <umairafzal@microsoft.com>
"""


# pylint: disable="W1202, W1203, W0621, W0613, W1514, W0612, W0622"

import json
import logging
import os
import time

import requests

logging.root.setLevel("INFO")
logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


# Response Code
successfulRespCodeOK = 200
successfulRespCode = 201
healthCheckPassed = "PASSED"
MainParameters = {}


class LandslideTC:
    """Handles all landslide test case call functions."""

    def __init__(self, tcDetails, nodeName, user_details, id=None):
        self.tcDetails = tcDetails
        self.id = id
        self.nodeName = nodeName
        self.user_details = user_details

    def startTc(self, tas_ip):

        postURL = "http://" + tas_ip + ":8080/api/runningTests/"

        # startResponse = requests.post('http://10.168.0.119:8080/api/runningTests/', data=self.tcDetails, auth=self.user_details)
        startResponse = requests.post(
            postURL, data=self.tcDetails, auth=self.user_details
        )
        parsedData = self.parseJSON(startResponse.json())

        logger.info(
            "\n Start {} - Parsed data - {}\n\n".format(self.nodeName, parsedData)
        )

        if startResponse.status_code == successfulRespCode:
            self.id = parsedData["id"]
        return startResponse.status_code

    def parseJSON(self, data):
        jsonData = json.dumps(data)
        parsedData = json.loads(jsonData)
        return parsedData

    def isTcHealthy(self, tas_ip):

        # healthCheckupURL =  "http://10.168.0.119:8080/api/runningTests/" + self.id + "/measurements/favorites"
        healthCheckupURL = (
            "http://"
            + tas_ip
            + ":8080/api/runningTests/"
            + self.id
            + "/measurements/favorites"
        )
        healthCheckResp = requests.get(
            healthCheckupURL, data=self.tcDetails, auth=self.user_details
        )
        parsedData = self.parseJSON(healthCheckResp.json())

        # logger.info('Health Check :  {} '.format(self.tcDetails))
        # logger.info('------------------------------------------')
        logger.info("\nHealth-Check Parsed Data  : {} \n\n".format(parsedData))

        criteriaNum = 1
        for element in parsedData["criteria"]:
            logger.info("Trial : {} ".format(element))
            criteriaNum += 1

        return parsedData["criteriaStatus"]

    def getTCStatus(self, tas_ip):
        # stausURL = "http://10.168.0.119:8080/api/runningTests/" + self.id
        stausURL = "http://" + tas_ip + ":8080/api/runningTests/" + self.id
        statusResp = requests.get(stausURL, data=self.tcDetails, auth=self.user_details)
        parsedData = self.parseJSON(statusResp.json())

        return parsedData["testStateOrStep"]

    # It will return Full State for Command Sequence TCs
    def getTcFullState(self, tas_ip):
        stausURL = "http://" + tas_ip + ":8080/api/runningTests/" + self.id
        statusResp = requests.get(stausURL, data=self.tcDetails, auth=self.user_details)
        parsedData = self.parseJSON(statusResp.json())

        fullState = parsedData["tsGroups"][0]["testCases"][0]["fullState"]

        return fullState

    # It will return parsedData for Command Sequence TCs
    def getTcCompleteStatus(self, tas_ip):
        stausURL = "http://" + tas_ip + ":8080/api/runningTests/" + self.id
        statusResp = requests.get(stausURL, data=self.tcDetails, auth=self.user_details)
        parsedData = self.parseJSON(statusResp.json())

        return parsedData

    def getFavoriteCounters(self, tas_ip):

        # healthCheckupURL =  "http://10.168.0.119:8080/api/runningTests/" + self.id + "/measurements/favorites"
        favoritesURL = (
            "http://"
            + tas_ip
            + ":8080/api/runningTests/"
            + self.id
            + "/measurements/favorites"
        )
        favoritesResp = requests.get(
            favoritesURL, data=self.tcDetails, auth=self.user_details
        )
        parsedData = self.parseJSON(favoritesResp.json())

        logger.info("\n Measurements Parsed Data  : {} \n\n".format(parsedData))

        favoriteCounters = {}
        for element in parsedData["favorites"]:
            key = list(element.keys())[0]
            value = list(element.values())[0]
            favoriteCounters[key] = value

        return favoriteCounters

    # It will continue the test from wait state
    def continueTc(self, tas_ip):

        postURL = (
            "http://"
            + tas_ip
            + ":8080/api/runningTests/"
            + self.id
            + "?action=continue"
        )

        # startResponse = requests.post('http://10.168.0.119:8080/api/runningTests/', data=self.tcDetails, auth=self.user_details)
        continueResponse = requests.post(
            postURL, data=self.tcDetails, auth=self.user_details
        )
        parsedData = self.parseJSON(continueResponse.json())

        logger.info(
            "\n Continue {} - Parsed data - {}\n\n".format(self.nodeName, parsedData)
        )

        return parsedData["result"]

    # /**** stopTC() - It will Stop Running TC ****/
    def stopTc(self, tas_ip):

        tcStatus = self.getTCStatus(tas_ip)
        logger.info(
            "\n Getting TC Status before stopping -\n {} \n\n\n".format(tcStatus)
        )

        if (
            (tcStatus.find("RUNNING") != -1)
            or (tcStatus.find("STARTED") != -1)
            or (tcStatus.find("Waiting") != -1)
        ):

            logger.info("TC Status is RUNNING/STARTED -- Trying to STOP")
            stopURL = (
                "http://"
                + tas_ip
                + ":8080/api/runningTests/"
                + self.id
                + "?action=stop"
            )

            stopResp = requests.post(
                stopURL, data=self.tcDetails, auth=self.user_details
            )
            parsedStopData = self.parseJSON(stopResp.json())

            logger.info("\n\n Stopping TC : {} ".format(self.tcDetails))
            logger.info("------------------------------------------")
            logger.info("ParsedStopData -- {}\n\n".format(parsedStopData))

            time.sleep(60)

            tcStatus = self.getTCStatus(tas_ip)
            retryCount = 0
            logger.info("TC Status :  {} ".format(tcStatus))

            # while (tcStatus != "COMPLETE") :
            while not tcStatus.startswith("COMPLETE"):
                time.sleep(30)
                tcStatus = self.getTCStatus(tas_ip)
                logger.info("{} : TC Status - {}".format(retryCount, tcStatus))
                retryCount = retryCount + 1
                if retryCount > 18:
                    break

    # Stop Command Sequenc TC and return execution results
    def stopCmdSeqTc_and_getResults(self, tas_ip):

        stopURL = (
            "http://" + tas_ip + ":8080/api/runningTests/" + self.id + "?action=stop"
        )
        stopResp = requests.post(stopURL, data=self.tcDetails, auth=self.user_details)
        parsedStopData = self.parseJSON(stopResp.json())
        logger.info("ParsedStopData -- {}\n\n".format(parsedStopData))
        time.sleep(5)

        tcStatus = ""
        tcResult = ""
        retryCount = 0

        while tcStatus != "COMPLETE":

            stausURL = "http://" + tas_ip + ":8080/api/runningTests/" + self.id
            statusResp = requests.get(
                stausURL, data=self.tcDetails, auth=self.user_details
            )
            parsedData = self.parseJSON(statusResp.json())

            tcStatus = parsedData["testStateOrStep"]
            tcResult = parsedData["criteriaStatus"]
            retryCount = retryCount + 1
            time.sleep(2)

            if retryCount > 900:
                break

        logger.info("\n TC-Status : {}".format(tcStatus))
        logger.info("\n TC-Result : {}".format(tcResult))

        return tcResult

    def getTcResult(self, tas_ip):
        # tcResultURL =  "http://10.168.0.119:8080/api/runningTests/" + self.id + "/measurements/favorites"
        tcResultURL = (
            "http://"
            + tas_ip
            + ":8080/api/runningTests/"
            + self.id
            + "/measurements/favorites"
        )
        response = requests.get(
            tcResultURL, data=self.tcDetails, auth=self.user_details
        )

        parsedData = self.parseJSON(response.json())
        logger.info("Kavita - get TC Result parsedData :  {} ".format(parsedData))
        return parsedData["criteriaStatus"]

    def getTcExecDetails(self, stat_dir, tas_ip):
        # tcResultURL =  "http://10.168.0.119:8080/api/runningTests/" + self.id + "/measurements/favorites"
        tcResultURL = (
            "http://"
            + tas_ip
            + ":8080/api/runningTests/"
            + self.id
            + "/measurements/favorites"
        )
        response = requests.get(
            tcResultURL, data=self.tcDetails, auth=self.user_details
        )

        parsedData = self.parseJSON(response.json())

        fileName = self.nodeName + "_stats"
        statFile = os.path.join(stat_dir, fileName)
        file = open(statFile, "a")

        file_content = "Statistics : {} \n\n".format(self.nodeName)

        favCount = 1
        for element in parsedData["favorites"]:
            file_content += "{} : {} \n".format(favCount, element)
            # logger.info(' * {}'.format(element))
            # logger.info(' * {}'.format(element))
            favCount += 1

        file.write(file_content)
        file.close()

    def getTcExecSummaryStats(self, tas_ip, stat_dir):
        tcResultURL = (
            "http://" + tas_ip + ":8080/api/runningTests/" + self.id + "/measurements/"
        )
        response = requests.get(
            tcResultURL, data=self.tcDetails, auth=self.user_details
        )

        parsedData = self.parseJSON(response.json())

        fileName = self.nodeName + "_stats"
        statFile = os.path.join(stat_dir, "Statistics", fileName)
        file = open(statFile, "a")

        file_content = ""

        file_content += (
            "Statistics validation status :\n" + "================================\n"
        )
        count = 1
        for element in parsedData["tabs"]["Pass-Fail History"]:
            file_content += "{} : {} - {}\n".format(
                count, element, parsedData["tabs"]["Pass-Fail History"][element]
            )
            count = count + 1

        file_content += "\n NGAP Statistics :\n" + "================================\n"
        count = 1
        for element in parsedData["tabs"]["NGAP"]:
            file_content += "{} : {} - {}\n".format(
                count, element, parsedData["tabs"]["NGAP"][element]
            )
            count = count + 1

        file_content += "\nMM Statistics :\n" + "================================\n"
        count = 1
        for element in parsedData["tabs"]["MM"]:
            file_content += "{} : {} - {}\n".format(
                count, element, parsedData["tabs"]["MM"][element]
            )
            count = count + 1

        file_content += "\nSM Statistics :\n" + "================================\n"
        count = 1
        for element in parsedData["tabs"]["SM"]:
            file_content += "{} : {} - {}\n".format(
                count, element, parsedData["tabs"]["SM"][element]
            )
            count = count + 1

        file.write(file_content)
        file.close()

    def execute_testSession_and_store_results(self, tas_ip, logdir, interface):
        response = self.startTc(tas_ip)

        if response != successfulRespCode:
            result = "FAILED"
            logger.info("\n Failed to Start TC - %s : %s", self.nodeName, response)
            return
        else:
            logger.info("\n Successfully Started TC - %s : %s", self.nodeName, response)

        time.sleep(20)

        # Check TC full state
        tcState = ""
        retryCount = 0

        # Edited below line from original script for tc state other than stopped...unsure if this is needed
        # while (tcState != 'STOPPED' and tcState != 'COMPLETE' and tcState != 'UNINITIALIZED' and tcState != 'STOPPING'):
        while (
            tcState != "STOPPED"
            and tcState != "COMPLETE"
            and tcState != "UNINITIALIZED"
            and tcState != "STOPPING"
        ):
            tcState = self.getTcFullState(tas_ip)

            if retryCount == 10 and tcState == "STARTED":
                logger.info(
                    "Failed to execute testcase aborting TC - %s", self.nodeName
                )
                break
            # increased sleep time
            logger.info("\n %s full-state - %s", self.nodeName, tcState)
            time.sleep(20)

            retryCount = retryCount + 1

        logger.info("\n %s full-state - %s", self.nodeName, tcState)

        # Stop TC and get TC Results
        result = self.stopCmdSeqTc_and_getResults(tas_ip)
        logger.info("---------- %s : %s", self.nodeName, result)

        # capture pcap
        self.downloadPcap(tas_ip, interface, logdir)

        # capture summary stats
        self.getTcExecSummaryStats(tas_ip, logdir)

        self.deleteTc(tas_ip)
        logger.info("\n Deleted TC")

        return result

    def deleteTc(self, tas_ip):
        # deleteURL = "http://10.168.0.119:8080/api/runningTests/" + self.id
        deleteURL = "http://" + tas_ip + ":8080/api/runningTests/" + self.id
        deleteResp = requests.delete(
            deleteURL, data=self.tcDetails, auth=self.user_details
        )

        parseDeleteData = self.parseJSON(deleteResp.json())
        logger.info("ParsedStopData -- {}\n\n".format(parseDeleteData))

    def downloadPcap(self, tas_ip, interface, logdir):

        logger.info("Download pcap - {}".format(self.nodeName))
        captureFileURL = (
            "http://"
            + tas_ip
            + ":8080/api/runningTests/"
            + self.id
            + "/captureFiles/0/"
            + interface
        )
        logger.info("Capture File URL  - {}".format(captureFileURL))
        response = requests.get(
            captureFileURL, data=self.tcDetails, auth=self.user_details
        )

        parsedStopData = self.parseJSON(response.json())
        logger.info("Save file pcap - {}".format(parsedStopData))

        if response.status_code == successfulRespCodeOK:
            parsedStopData = self.parseJSON(response.json())

            pcapURL = parsedStopData["captureFileUrl"]

            r = requests.get(pcapURL, allow_redirects=True)
            name = self.nodeName + ".pcap"
            pcapFile = os.path.join(logdir, "PCAPs", name)
            file = open(pcapFile, "wb")
            file.write(r.content)
            file.close()

    def get_all_landslide_measurements(self, tas_ip):
        tc_result_url = f"http://{tas_ip}:8080/api/runningTests/{self.id}/measurements"
        response = requests.get(
            tc_result_url, data=self.tcDetails, auth=self.user_details
        )
        return self.parseJSON(response.json())

    def get_landslide_measurement(self, tas_ip, tab_name, measurement=""):
        """
        This method returns particular measurement value if measurement name is provided, else full tab measurements are returned.

        Args:
            tas_ip (str): Landslide TAS IP
            tab_name (str): Tab name inside the Session Reports section in Landslide
            measurement (str, ooptional): an element inside a tab. For example, "Attempted Session Connects".
                if no measurement is provided, the whole tab is returned.

        Returns:
            str: The 'measurement' in the specified 'tab_name' OR
            list: A list of all measurements in the specified 'tab_name'

        Sample tabs from Landslide Report:
            "tabs": {
                "Pass-Fail History": {
                    "C0 Session_check": "PENDING",
                    ...
                },
                "Test Summary": {
                    "Start Time": "Thu Sep 09 22:22:15 CDT 2021",
                    "Attempted Session Connects": "4000",
                    "Total Data Sent + Received Packets/Sec  (P-I)": "34,247.933",
                    ...
                },
                "L5-7 Client|Basic": {
                    "Basic Data Message Minimum RTD In-Interval": "1 us",
                    "Basic Data Message Maximum RTD In-Interval": "6.46 ms",
                    ...
                }
            }
        """
        landslide_stats = self.get_all_landslide_measurements(tas_ip)

        if measurement:
            logger.info(
                f"Fetching {measurement} from tab {tab_name} for Test ID {self.id}"
            )
            measurement_value = (
                landslide_stats.get("tabs").get(tab_name).get(measurement)
            )
            logger.info(f"{measurement} = {measurement_value}")
            return measurement_value

        else:
            logger.info(f"Fetching tab {tab_name} for Test ID {self.id}")
            tab_measurements = landslide_stats.get("tabs").get(tab_name)
            logger.info(f"Measurements inside {tab_name} = {tab_measurements}")
            return tab_measurements
