#!/usr/bin/env python3
# Copyright 2022 Metaswitch Networks - Highly Confidential Material
"""Contains resources to update nipap when ips are booked internally to the Landslide test runner."""

import logging
import os
from tokenize import String
from typing import List
from xmlrpc.client import Boolean

import pynipap
from pynipap import Prefix, Tag

# Query to show an IP is already reserved for use in the LTR
is_ltr_ip = {
    "operator": "equals_any",
    "val1": "tags",
    "val2": "landslide-test-runner",
}

# Query to show an IP is reserved for automated activation, with description landslide-test-runner-<number>
is_auto_ip = {
    "operator": "regex_match",
    "val1": "description",
    "val2": "^ngc-ts1-landslide-test-runner-auto",
}


def get_range_query(ip_range="10.232.44.0/24"):
    """
    Get a query that searches in a given range.

    Default is 10.232.44.0/24.
    """
    return {
        "operator": "contained_within",
        "val1": "prefix",
        "val2": ip_range,
    }


def get_is_tagged_query(tag: String):
    """Get a query that searches for a given tag."""
    return {"operator": "equals_any", "val1": "tags", "val2": tag}


def get_ip_query(ip: String):
    """Get a query that searches for a given ip."""
    return {"operator": "equals", "val1": "prefix", "val2": ip}


def combine_queries(queries: list):
    """
    Takes a list of queries, and combines them into one query for searching nipap.

    This is a recursive method, and will eventually return a dictionary.
    """
    len(queries)
    if len(queries) == 1:
        return queries[0]
    else:
        q1 = queries.pop()
        q2 = queries.pop()
        q = {
            "operator": "and",
            "val1": q1,
            "val2": q2,
        }
        queries.append(q)
        return combine_queries(queries)


def get_tagged_prefixes(tag: String) -> list:
    """Get prefixes with a given tag."""
    query = combine_queries(
        [get_is_tagged_query(tag), get_range_query(), is_auto_ip, is_ltr_ip]
    )
    return get_prefixes(query)


def setup_nipap_authentication() -> None:
    """
    Setup all of the global variables for connecting to nipap.

    This is necessary to run most of the following functions.
    """
    username = os.environ["NIPAP_USERNAME"]
    password = os.environ["NIPAP_PASSWORD"]
    pynipap.xmlrpc_uri = (
        "http://" + username + ":" + password + "@10.231.137.79:1337/XMLRPC"
    )
    pynipap.AuthOptions({"authoritative_source": "nipap"})


def activate_nipap_ip(user: String, ip: String) -> None:
    """Given an IP reserved for the LTR, changes the tags of the prefix to active."""
    setup_nipap_authentication()
    prefix = get_prefix_with_ip(ip)
    tags = prefix.tags.keys()

    if "landslide-test-runner" not in tags:
        raise Exception("IP is not designated for use by the Landslide Test Runner")
    elif "available" not in tags:
        raise Exception("IP is currently in use.")

    edit_tag(prefix, "available", "active-" + user)
    prefix.save()
    return


def auto_activate_nipap_ip(user: String) -> Prefix:
    """Find a prefix available for automated booking, and change its tags to active."""
    setup_nipap_authentication()
    available_prefixes = get_tagged_prefixes("available")

    if len(available_prefixes) == 0:
        raise Exception("All prefixes are in use")

    prefix = available_prefixes[0]
    edit_tag(prefix, "available", "active-" + user)
    prefix.save()
    return prefix.prefix


def deactivate_nipap_ip(ip: String, user: String) -> Boolean:
    """Make an active prefix available for automated booking."""
    setup_nipap_authentication()
    prefix = get_prefix_with_ip(ip)

    if not edit_tag(prefix, "active-" + user, "available"):
        logging.error("Tag is not used for automated booking, or user is incorrect.")
        return False

    prefix.save()
    return True


def deactivate_all_ips(user: String):
    """Deactivate all ips belonging to a single user."""
    logging.info("Cleaning up IPs...")
    prefix_list = get_tagged_prefixes("active-" + user)
    for prefix in prefix_list:
        if deactivate_nipap_ip(prefix.prefix, user):
            logging.info("Deactivated IP: %s", prefix.prefix)


def edit_tag(prefix: Prefix, original_tag: String, desired_tag: String) -> Boolean:
    """Edit a tag of a prefix."""
    tag_dict = prefix.tags

    tag = tag_dict.pop(original_tag, "")  # Return empty string if nothing to pop()
    if tag == "":  # if the tag doesn't exist, return false
        logging.error("Original tag does not exist")
        return False

    tag.name = desired_tag
    tag_dict[desired_tag] = tag
    prefix.tags = tag_dict
    return True


def get_prefixes(query: dict) -> List:
    """Get a list of prefixes that satisfy the given query."""
    setup_nipap_authentication()
    search_result = Prefix.smart_search("", {}, query)
    prefixes = search_result["result"]
    return prefixes


def get_prefix_with_ip(ip: String) -> Prefix:
    """Get a prefix with a given ip."""
    query = get_ip_query(ip)
    prefix_list = get_prefixes(query)

    if len(prefix_list) == 0:
        raise Exception("IP given does not belong to a valid prefix")

    return prefix_list[0]


def check_prefix_name(prefix: Prefix, name=None) -> Prefix:
    """Check a prefix has a given name."""
    if prefix.description != name:
        raise Exception("Given IP has name %s, not %s" % (prefix.description, name))


def reserve_ip(
    ip: String, user: String, current_prefix_name=None, personal_use=False
) -> None:
    """
    Reserves a specific ip for use in the landslide test runner, either for automated booking (default) or for personal use.

    If a name is already assigned to the prefix, it must be given as a parameter.
    """
    prefix = get_prefix_with_ip(ip)
    check_prefix_name(prefix, current_prefix_name)

    if personal_use:
        prefix.description = "ngc-ts1-landslide-test-runner-" + user
    else:
        prefix.description = "ngc-ts1-landslide-test-runner-auto"

    prefix.tags = {
        "landslide-test-runner": Tag.from_dict({"name": "landslide-test-runner"}),
        "available": Tag.from_dict({"name": "available"}),
    }

    prefix.save()
    return


def delete_custom_ip(ip: String, user: String) -> None:
    """
    Deletes a prefix created via reserve_ip.

    If deleting an auto IP, set user = auto.
    """
    prefix = get_prefix_with_ip(ip)
    check_prefix_name(prefix, "ngc-ts1-landslide-test-runner-" + user)

    prefix.description = None
    prefix.tags = {}
    prefix.save()
    return
