#!/usr/bin/env python3
# Copyright 2022 Metaswitch Networks - Highly Confidential Material
# pylint: disable=E1101

"""File to load and run landslide tests based on input from a yaml file."""

import argparse
import base64
import json
import logging
import os
from dataclasses import dataclass, field
from typing import List, Optional

import requests
import yaml
from dacite import from_dict

from landslide_test_runner.json_map import JSON_MAP
from landslide_test_runner.update_nipap import auto_activate_nipap_ip

# pylint: disable=E1101 # override dataclass field error


def add_to_landslide_json(landslide_json: dict, key: str, value: str):
    """
    Adds a key-value pair to the JSON to send to Landslide.

    Uses the jsonpointer found in JSON_MAP.
    """
    if key not in JSON_MAP.keys():
        logging.warning("unrecognised key: %s", key)
    else:
        pointer = JSON_MAP[key]
        pointer.set(landslide_json, value)

    return


def landslide_api(request_type: str, url: str, payload=None):
    """Make a request using the Landslide API."""
    headers = {
        "user-agent": os.environ["LANDSLIDE_USER_AGENT"],
        "authorization": os.environ["LANDSLIDE_AUTHORIZATION"],
    }

    if request_type is None:
        r = requests.request(request_type, url, headers=headers, verify=False)
    else:
        r = requests.request(
            request_type, url, headers=headers, verify=False, data=payload
        )

    if r.status_code not in range(200, 299):
        print(r.json())
        raise Exception(
            "Error in request to Landslide. Status code:" + str(r.status_code)
        )

    return r.json()


def parse_arguments(args=None):
    """Pass yaml file as input for test configuration."""
    parser = argparse.ArgumentParser(
        description="Spirent Test Runner for Fusion Core VM Instances"
    )

    parser.add_argument(
        "test_config",
        nargs="?",
        action="store",
        default="example_parameters.yaml",
        help="Configuration yaml file for test session being run",
    )
    parser.add_argument(
        "-t",
        dest="start",
        action="store_true",
        help="Start test and dont wait",
    )

    parser.add_argument(
        "-m",
        dest="monitor",
        action="store_true",
        help="Don't start test, just monitor it",
    )

    parser.add_argument(
        "-s", dest="stop", action="store_true", help="Stop test if running"
    )

    args = parser.parse_args(args)
    return args


def get_config(config_filename: str):
    """Parse yaml input file and return dataclass instances."""
    data_file = open(config_filename, "r", encoding="utf8").read()
    data = yaml.safe_load(data_file)
    return from_dict(data_class=TestConfig, data=data)


def get_session_id(session_name, tas_ip: str) -> str:
    """Get running test and return id."""
    url = f"http://{tas_ip}:8080/api/runningTests/"
    r = landslide_api("GET", url)

    for inst in r["runningTests"]:
        if inst["name"] == session_name:
            session_id = str(inst["id"])
            logging.info("Session found with ID %s", session_id)
            return session_id
        else:
            return "No session with the given name found."


@dataclass
class Parameters:
    """
    Modify any extra (optional) test case details.

    If not needed, Parameters might not be present in the yaml input.
    """

    subscribers: Optional[str] = None
    pdu_sessions: Optional[str] = None
    activation_rate: Optional[str] = None
    deactivation_rate: Optional[str] = None

    def as_json(self):
        """Constructs a json for Landslide that represents our parameters."""
        parameter_json = {
            "GnbControlAddr": {},
            "NetworkHostAddrLocal": {},
        }

        for key in self.__dataclass_fields__.keys():
            parameter_value = getattr(self, key)
            if parameter_value is None:
                continue
            add_to_landslide_json(parameter_json, key, parameter_value)

        return parameter_json


@dataclass
class TestCase:
    """Modify the test case details."""

    test_case_name: str
    parameters: Optional["Parameters"] = from_dict(data_class=Parameters, data={})
    gnb_n2: Optional[str] = None
    gnb_n6: Optional[str] = None

    def check_test_case(self, user: str) -> None:
        """Runs checks on the input case, assigns fields where necessary."""
        # Check that the GnB details are correct, assigns IPs if necessary.
        if self.gnb_n2 is None:
            self.gnb_n2 = auto_activate_nipap_ip(user)
            logging.info("No input n2 GnB - assigned IP: %s", self.gnb_n2)
        if self.gnb_n6 is None:
            self.gnb_n6 = auto_activate_nipap_ip(user)
            logging.info("No input n6 GnB - assigned IP: %s", self.gnb_n6)
        return

    def as_json(self, sut) -> dict:
        """Constructs a json for Landslide that represents a single test case."""
        # Compile all of our optional parameters.
        parameter_json = self.parameters.as_json()

        # Add the GnB address twice, to two different spots in the json.
        add_to_landslide_json(parameter_json, "gnb_n2", self.gnb_n2)
        add_to_landslide_json(parameter_json, "gnb_n6", self.gnb_n6)

        if sut.nat_enabled:  # if nat is enabled, we don't need a next hop ip.
            add_to_landslide_json(parameter_json, "nat_enabled", "true")
        else:
            add_to_landslide_json(parameter_json, "sut_n6", sut.n6)
            add_to_landslide_json(parameter_json, "sut_n6_interface", "eth1")
            add_to_landslide_json(parameter_json, "nat_enabled", "false")

        return {"name": self.test_case_name, "parameters": parameter_json}


@dataclass
class TestSession:
    """Test Session containing list of test case instances to be run."""

    test_session_name: str
    test_folder_name: str
    test_case_list: Optional[List[TestCase]] = field(default_factory=lambda: [])
    duration: Optional[int] = None

    test_folder_id: Optional[
        int
    ] = None  # assigned from the test_folder_name in run_checks

    def check_test_session(self, user: str, tas_ip: str) -> None:
        """Runs checks on the input session, assigns fields where necessary."""
        self.assign_test_folder_id(tas_ip)
        self.check_test_case_list(user, tas_ip)
        return

    def assign_test_folder_id(self, tas_ip: str):
        """Finds the test folder ID that corresponds to the name provided, and assigns it to test_folder_name."""
        r = landslide_api("GET", "http://" + tas_ip + ":8080/api/libraries/")
        for library in r["libraries"]:
            if library["name"] == self.test_folder_name:
                self.test_folder_id = library["id"]
                return
        raise Exception("Cannot find test folder " + self.test_folder_name)

    def check_test_case_list(self, user: str, tas_ip: str):
        """Checks that the number of test cases is correct, and that the correct names are given."""
        url = (
            "http://"
            + tas_ip
            + ":8080/api/libraries/"
            + str(self.test_folder_id)
            + "/testSessions/"
            + self.test_session_name
        )
        r = landslide_api("GET", url)
        pointer = JSON_MAP["test_case_list"]
        test_case_list = pointer.get(r)

        n_test_cases = len(test_case_list)
        if n_test_cases != len(self.test_case_list) and n_test_cases != 0:
            raise Exception(
                str(n_test_cases)
                + " test cases listed in session "
                + self.test_session_name
                + ", "
                + str(len(self.test_case_list))
                + " listed in Landslide."
            )

        for tc in self.test_case_list:
            tc.check_test_case(user)

    def as_json(self, sut) -> dict:
        """Constructs a json for Landslide that represents a single test session."""
        ts_json = {"tsGroups": [{}]}  # blank dict for ts parameters

        for key in self.__dataclass_fields__.keys():
            value = getattr(self, key)

            # If key has no value, skip.
            if value is None:
                continue

            # We only use this to get the test folder ID
            if key == "test_folder_name":
                continue

            # If key is "test_cases", add every test case in the list to the session.
            elif key == "test_case_list":
                test_case_list = value
                tc_json_list = []
                for tc in test_case_list:
                    tc_json_list.append(tc.as_json(sut))
                add_to_landslide_json(
                    ts_json, key, tc_json_list
                )  # add the test case jsons to our test session json

            else:
                add_to_landslide_json(ts_json, key, value)

        return ts_json


@dataclass
class TestConfig:
    """Overarching data class to store all of the test configuration subclasses."""

    identifier: Optional[str]
    spirent_configuration: "SpirentConfiguration"
    sut: "Sut"
    test_session_list: List[TestSession]
    test_server: Optional["TestServer"] = None

    def check_test_config(self, user: str) -> None:
        """Runs checks on the input config, assigns fields where necessary."""
        logging.info("Running checks...")
        tas_ip = self.spirent_configuration.tas_ip  # used in a few checks

        for ts in self.test_session_list:
            ts.check_test_session(user, tas_ip)
        self.spirent_configuration.check_spirent_configuration()
        sut_created = self.sut.check_sut(tas_ip)

        return sut_created


@dataclass
class SpirentConfiguration:
    """Class to configure spirent login details and the tas to run on."""

    username: Optional[str]
    password: Optional[str]
    tas_ip: Optional[str]

    def check_spirent_configuration(self):
        """
        If username or password is missing, use the defaults.

        Same for tas_ip.
        """
        if self.username is None or self.password is None:
            logging.error("Spirent username or password is missing- using defaults.")
            self.username = "stability"
            self.password = base64.b64decode("U3RhYjFsMXR5").decode("ascii")
        if self.tas_ip is None:
            logging.error("TAS IP is missing- using default.")
            self.tas_ip = "10.232.27.50"

    def get_login_details(self):
        """Get login details in a good format for running tests."""
        return (self.username, self.password)


@dataclass
class TestServer:
    """Test Server id to use for running tests and any updates to the test server."""

    id: str

    def check_test_server(self):
        """Check that the test server details are correct, and that IPs are in the correct range."""
        return


@dataclass
class Sut:
    """SUT class to create and manage existing SUTs in the Landslide test server."""

    name: str
    nat_enabled: bool
    n2: Optional[str] = None  # Optional if SUT already exists
    n6: Optional[str] = None  # If NAT is enabled, n6 isn't needed.

    id: Optional[int] = None  # Added during check_sut()

    def get_spirent_suts(self, tas_ip) -> list:
        """Gets a list of suts stored on Spirent."""
        r = landslide_api("GET", "http://" + tas_ip + ":8080/api/suts")
        return r["suts"]

    def create_spirent_sut(self, tas_ip) -> None:
        """Adds a sut to Spirent with our details."""
        if self.n2 is None:
            raise Exception("When creating a new SUT, n2 must be provided.")

        payload = json.dumps(
            {"ip": self.n2, "name": self.name, "managementIp": "1.1.1.1"}
        )
        landslide_api("POST", "http://" + tas_ip + ":8080/api/suts/", payload=payload)

    def assign_id_if_sut_exists(self, tas_ip) -> bool:
        """
        Checks if our name is assigned to a SUT already.

        If it is, sets our ID to the correct value.
        """
        suts = self.get_spirent_suts(tas_ip)
        for sut in suts:
            if self.name == sut["name"]:
                self.id = sut["id"]
                return True

        return False

    def assign_free_id(self, tas_ip) -> bool:
        """Finds an ID not currently in use, and sets our ID to this value."""
        suts = self.get_spirent_suts(tas_ip)
        for i in range(1, 100):
            for sut in suts:
                if i == sut["id"]:
                    break

            self.id = i
            return True

    def check_sut(self, tas_ip) -> bool:
        """Check that the SUT details are correct, create an SUT if necessary."""
        if self.assign_id_if_sut_exists(tas_ip):
            logging.info("Using existing SUT %s with ID %s.", self.name, self.id)
            sut_created = False
        elif self.assign_free_id(tas_ip):
            self.create_spirent_sut(tas_ip)
            logging.info(
                "No sut with given name. Created SUT %s with ID %s.", self.name, self.id
            )
            sut_created = True
        else:
            raise Exception("No SUT IDs in range 1-100 are available.")

        if not self.nat_enabled and self.n6 is None:
            raise Exception("NAT is disabled, an n6 address must be provided.")

        return sut_created

    def delete_sut(self, tas_ip):
        """Delete the SUT with our ID."""
        logging.info("Deleting SUT...")
        landslide_api(
            "DELETE", url="http://" + tas_ip + ":8080/api/suts/" + str(self.id)
        )
        return
